static char rcsid[] = "@(#)$Id: pgrng_bfr_binding.c,v 2.3 2014/07/14 15:07:27 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.3 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elm_defs.h"
#include "pg_range_imp.h"

DEBUG_VAR(DebugIO,__FILE__,"stateio");

#define PGRNG_BFR_magic	     0xEB0F

struct pgrng_bfr_binding {
    unsigned short              magic; /* PGRNG_BFR_magic */

    int refcount;

    /* target's refcount is NOT incremented: */

    struct pager_range  *range;
    struct stringbuffer *buffer;
    int                  buffer_index;

    /* range and buffer also count as refcount */

};

/* 1) binding -> pager_range -------------------------------------------- */

static void clear_pgrng_bfr_binding P_((struct pgrng_bfr_binding **binding));

void pgrng_bfr_bnd_clear_range(binding,range)
     struct pgrng_bfr_binding **binding;
     const struct pager_range *range;
{
    if (PGRNG_BFR_magic != (*binding)->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "pgrng_bfr_bnd_clear_range",
	      "Bad magic number",0);

    if (range != (*binding)->range)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "pgrng_bfr_bnd_clear_range",
	      "Link (range) failure",0);
    (*binding)->range = NULL;
    
    clear_pgrng_bfr_binding(binding);
}

void pgrng_bfr_bnd_set_range(binding,range)
     struct pgrng_bfr_binding *binding;
     struct pager_range *range;
{
    if (PGRNG_BFR_magic != binding->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "pgrng_bfr_bnd_set_range",
	      "Bad magic number",0);

    if (binding->range)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "pgrng_bfr_bnd_set_range",
	      "Link (range) failure",0);

    binding->range = range;
}

/* Caller must call free_pager_range() */
struct pager_range * pgrng_bfr_bnd_get_range(binding)
     const struct pgrng_bfr_binding *binding;
{
    if (PGRNG_BFR_magic != binding->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "pgrng_bfr_bnd_get_range",
	      "Bad magic number",0);

    if (binding->range)
	inc_pager_range_refcount(binding->range);

    return binding->range;
}

/* 2) binding -> stringbuffer ------------------------------------------- */

void pgrng_bfr_bnd_clear_stringbuffer(binding,buffer,index)
     struct pgrng_bfr_binding **binding;
     const struct stringbuffer *buffer;
     int index;
{
    if (PGRNG_BFR_magic != (*binding)->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "pgrng_bfr_bnd_clear_stringbuffer",
	      "Bad magic number",0);

    if (buffer != (*binding)->buffer ||
	index  != (*binding)->buffer_index)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "pgrng_bfr_bnd_clear_stringbuffer",
	      "Link (stringbuffer) failure",0);	
    (*binding)->buffer = NULL;
    (*binding)->buffer_index = -1;

    clear_pgrng_bfr_binding(binding);
}

void pgrng_bfr_bnd_set_stringbuffer(binding,buffer,index)
     struct pgrng_bfr_binding *binding;
     struct stringbuffer *buffer;
     int index;
{
    if (PGRNG_BFR_magic != binding->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "pgrng_bfr_bnd_clear_stringbuffer",
	      "Bad magic number",0);

    if (binding->buffer ||
	-1 != binding->buffer_index)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "pgrng_bfr_bnd_set_stringbuffer",
	      "Link (stringbuffer) failure",0);	

    binding->buffer       = buffer;
    binding->buffer_index = index;
}

/* Caller must call free_stringbuffer() */
struct stringbuffer * pgrng_bfr_bnd_get_stringbuffer(binding,index)
     struct pgrng_bfr_binding *binding; 
     int *index;
{
    if (PGRNG_BFR_magic != binding->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "pgrng_bfr_bnd_get_stringbuffer",
	      "Bad magic number",0);

    if (index)
	*index = binding->buffer_index;

    if (binding->buffer)
	inc_stringbuffer_refcount(binding->buffer);

    return binding->buffer;
}


/* 3) general ---------------------------------------------------------- */

struct pgrng_bfr_binding * new_pgrng_bfr_binding() 
{
    struct pgrng_bfr_binding *ret = safe_malloc(sizeof (*ret));

    /* bzero is defined hdrs/elm_defs.h */
    bzero((void *)ret,sizeof (*ret));

    ret->magic = PGRNG_BFR_magic;
    ret->refcount = 1;
    
    ret->range = NULL;
    ret->buffer = NULL;
    ret->buffer_index = -1;

    return ret;
}


void free_pgrng_bfr_binding(binding)
     struct pgrng_bfr_binding **binding;
{

    if (PGRNG_BFR_magic != (*binding)->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "free_pgrng_bfr_binding",
	      "Bad magic number",0);

    (*binding)->refcount--;

    clear_pgrng_bfr_binding(binding);
}


void inc_pgrng_bfr_binding_refcount(binding)
     struct pgrng_bfr_binding *binding;
{
    if (PGRNG_BFR_magic != binding->magic)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "inc_pgrng_bfr_binding_refcount",
	      "Bad magic number",0);

    binding->refcount++;
}

static void clear_pgrng_bfr_binding(binding)
     struct pgrng_bfr_binding **binding;
{
    if ((*binding)->refcount < 0)
	panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "clear_pgrng_bfr_binding",
	      "Bad recount",0);

    if (((*binding)->buffer   && -1 == (*binding)->buffer_index) ||
	(! (*binding)->buffer && -1 != (*binding)->buffer_index))
		panic("PAGER RANGE PANIC",__FILE__,__LINE__,
	      "clear_pgrng_bfr_binding",
	      "Bad buffer_index",0);
	
    if ((*binding)->refcount ||
	(*binding)->range ||
	(*binding)->buffer) 
	return;

    (*binding)->magic = 0;  /* Invalidate */
    free(*binding);
    *binding = NULL;
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
